
#### Interface

EL.LogLikelihood <- function(X, Y, d, t, type="roc", bw.fun=bw.nrd0, bw.fixed=NULL)
  {
    EL.local(ELLR, X, Y, d, t, method=type, bw.exp=NULL, bw.fun=bw.fun, bw.fixed=bw.fixed)
  }

EL.curve <- function(X, Y, t, type="roc",
                     conf.level=NULL,
                     bw.fun=bw.nrd0, bw.fixed=NULL, 
                     sim.conf.level=NULL, sim.numboot=200,
                     bw.exp=NULL, ...)
  {
    mm <- function(tt)
      {
        d <- sapply(tt, function(t) deltasolve(X, Y, t))
        if (is.null(conf.level))
          ci <- NA
        else
          ci <- sapply(1:length(tt), function(k) ELconf(X, Y, tt[k], conf.level, delta=d[k]))

        if (is.null(sim.conf.level))
          {
            si <- NA
            bs <- NA
          }
        else
          {
            errnum <- 0
            bootup <- function()
              {
                X1 <- sample(X, replace=TRUE)
                Y1 <- sample(Y, replace=TRUE)
                tryCatch(max(mapply(ELLR, d, tt, MoreArgs=list(X=X1, Y=Y1))),
                         error=function(e)
                         {
                           errnum <- errnum + 1
                           if (errnum > 1000)
                             stop("Over 1000 errors encountered bootstraping")
                           warning("Error bootstraping sample")
                           bootup()
                         })
              }
            bs <- sort(replicate(sim.numboot, bootup()))
            bs <- bs[ceiling(sim.conf.level * sim.numboot)]
            print(bs)
            si <- sapply(1:length(tt), function(k) ELconf(X, Y, tt[k], pchisq(bs, 1), delta=d[k]))
          }
        
        list(estim=d, conf.int=ci, boot.crit.val=bs, simult.conf.int=si)
      }
    EL.local(mm, t, method=type, bw.exp=bw.exp, bw.fun=bw.fun, bw.fixed=bw.fixed)
  }

EL.means.p <- function(X, Y, d=0)
  {
    1 - pchisq(EL.local(ELLR, X, Y, d, 0), 1)
  }

EL.means <- function(X, Y, conf.level=0.95)
  {
    mm <- function()
      {
        d <- deltasolve(X, Y, 0)
        ci <- ELconf(X, Y, 0, conf.level, delta=d)
        list(estim=d, conf.int=ci)
      }
    EL.local(mm)
  }


#### Globals setup

set.globals <- function()
  {
    assign("g.lambda1", 0, envir=globalenv())
    assign("g.lambda2", 0, envir=globalenv())
    assign("g.wval1", NULL, envir=globalenv())
    assign("g.wval2", NULL, envir=globalenv())
    assign("g.lambdarange1", 0, envir=globalenv())
    assign("g.lambdarange2", 0, envir=globalenv())
    assign("g.degenerate", FALSE, envir=globalenv())
    assign("g.precision", 1e-2, envir=globalenv())
  }


#### Bandwidth setup

set.bw <- function(type="e", arg=-0.2)
  {
    if (!(type %in% c("exponent", "function", "manual", "e", "f", "m")))
      stop(paste("Unknown bandwidth method: ", type, "\n",
                 "Recognized bandwidth methods are: function, exponent, manual."))

    assign("bw.X", 0, envir=globalenv())
    assign("bw.Y", 0, envir=globalenv())
    assign("bw.type", type, envir=globalenv())
    assign("bw.arg", arg, envir=globalenv())
    assign("initbw", bw.init, envir=globalenv())
  }

bw.init <- function(X, Y)
  {
    if (bw.type %in% c("manual", "m"))
      {
        bw.X <<- bw.arg[1]
        bw.Y <<- bw.arg[2]
        return()
      }
    bw <- function(X)
      {
        if (bw.type %in% c("exponent", "e"))
          length(X)^bw.arg
        else
          if (bw.type %in% c("function", "f"))
            bw.arg(X)
      }
    if (!use.smoothing)
      return()
    if (bw.X == 0)
      bw.X <<- bw(X)
    if (bw.Y == 0)
      bw.Y <<- bw(Y)
  }

#### Kernel setup

set.kernel <- function()
  {
    assign("krnl", kernel.norm, envir=globalenv())
    assign("dkrnl", kernel.dnorm, envir=globalenv())
    assign("ikrnl", kernel.inorm, envir=globalenv())
  }

kernel.norm <- function(t)
	pnorm(t)

kernel.dnorm <- function(t)
	dnorm(t)

kernel.inorm <- function(t)
	qnorm(t)

kernel.fun <- function(t)
  {
    res <- 0.9375 * (t - 2/3 * t ^ 3 + t^5/5) + 0.5
    res[res < 0] <- 0
    res[res > 1] <- 1
    res
  }

kernel.dfun <- function(t)
  {
    res <- 1 - t^2
    res[res < 0] <- 0
    0.9375 * res^2
  }

kernel.inv <- function(t)
  {
    dkern <- function(x)
      krnl(x) - t
    uniroot(dkern, c(-1, 1))$root
  }

#### Method setup

set.method <- function(type)
  {
    if (!(type %in% c("pp", "qq", "roc", "mean", "fdiff", "qdiff")))
      stop(paste("Unknown EL method: ", type, "\n",
                 "Recognized EL methods are: pp, qq, roc, mean, fdiff, qdiff."))
    

    assign("w1", get(paste("w1.", type, sep="")), envir=globalenv())
    assign("w2", get(paste("w2.", type, sep="")), envir=globalenv())
    assign("alpha1", get(paste("alpha1.", type, sep="")), envir=globalenv())
    assign("alpha2", get(paste("alpha2.", type, sep="")), envir=globalenv())
    assign("trange", get(paste("trange.", type, sep="")), envir=globalenv())
    assign("deltarange", get(paste("deltarange.", type, sep="")), envir=globalenv())
    assign("thetarange", get(paste("thetarange.", type, sep="")), envir=globalenv())
    assign("use.smoothing", get(paste("use.smoothing.", type, sep="")), envir=globalenv())    
  }


### Mean
w1.mean <- function(X, theta, delta, t)
  X - theta

w2.mean <- function(X, theta, delta, t)
  X - theta + delta

alpha1.mean <- function(X, theta, delta, t)
  -1

alpha2.mean <- function(X, theta, delta, t)
  -1

trange.mean <- function(X, Y)
  stop("Mean differences are not dependent on the parameter t.")

deltarange.mean <- function(X, Y, t)
  {
    tr <- c(min(X), max(X))
    c(tr[1] -max(Y) + g.precision, tr[2]-min(Y)- g.precision)
  }

thetarange.mean <- function(X, Y, delta, t)
  {
    tr <- c(min(X) + g.precision, max(X) - g.precision)
    dr <- c(min(Y) + delta, max(Y) + delta)
    c(max(tr[1], dr[1]-g.precision), min(tr[2], dr[2]+g.precision))
  }

use.smoothing.mean <- FALSE

### ROC

w1.roc <- function(X, theta, delta, t)
  krnl((theta - X) / bw.X) - (1 - delta)

w2.roc <- function(X, theta, delta, t)
  krnl((theta - X) / bw.Y) - (1 - t)

alpha1.roc <- function(X, theta, delta, t)
  dkrnl((theta - X) / bw.X) / bw.X

alpha2.roc <- function(X, theta, delta, t)
  dkrnl((theta - X) / bw.Y) / bw.Y

trange.roc <- function(X, Y)
  {
    c(0.01, 0.99)
  }

deltarange.roc <- function(X, Y, t)
  {
    c(g.precision, 1 - g.precision)
  }

thetarange.roc <- function(X, Y, delta, t)
  {
    XX <- sort(X)
    YY <- sort(Y)
    nX <- length(X)
    nY <- length(Y)
    dX <- bw.X * ikrnl(delta)
    dY <- bw.Y * ikrnl(t)
    c(max(XX[2] - dX, YY[2] - dY) + g.precision, min(XX[nX - 1] - dX, YY[nX - 1] - dY) - g.precision)
  }

use.smoothing.roc <- TRUE


### P-P plots


w1.pp <- function(X, theta, delta, t)
  krnl((theta - X) / bw.X) - delta

w2.pp <- function(X, theta, delta, t)
  krnl((theta - X) / bw.Y) - t

alpha1.pp <- function(X, theta, delta, t)
  dkrnl((theta - X) / bw.X) / bw.X

alpha2.pp <- function(X, theta, delta, t)
  dkrnl((theta - X) / bw.Y) / bw.Y

trange.pp <- function(X, Y)
  {
    c(0.01, 0.99)
  }

deltarange.pp <- function(X, Y, t)
  {
    dY <- bw.Y * ikrnl(t)
    tr <- c(min(Y) + dY, max(Y) + dY)
    c(krnl((tr[1]-max(X)) / bw.X) + g.precision, krnl((tr[2]-min(X)) / bw.X) - g.precision)
  }

thetarange.pp <- function(X, Y, delta, t)
  {
    dX <- bw.X * ikrnl(delta)
    dY <- bw.Y * ikrnl(t)
    dd <- c(max(min(X) + dX, min(Y) + dY), min(max(X) + dX, max(Y) + dY))
    c(dd[1] + 0.01 * (dd[2] - dd[1]), dd[2] - 0.01 * (dd[2] - dd[1]))
  }

use.smoothing.pp <- TRUE

### Q-Q plots
w1.qq <- function(X, theta, delta, t)
  krnl((delta - X) / bw.X) - theta

w2.qq <- function(X, theta, delta, t)
  krnl((t - X) / bw.Y) - theta

alpha1.qq <- function(X, theta, delta, t)
  -1

alpha2.qq <- function(X, theta, delta, t)
  -1

trange.qq <- function(X, Y)
  c(min(Y), max(Y))

deltarange.qq <- function(X, Y, t)
  {
    c(min(X) + g.precision, max(X) - g.precision)
  }
    
thetarange.qq <- function(X, Y, delta, t)
  {
    ww <- w2.qq(Y, 0, 0, t)
    tr <- c(min(ww), max(ww))
    ww <- w1.qq(X, 0, delta, 0)
    dr <- c(min(ww), max(ww))
    c(max(tr[1], dr[1]) + g.precision, min(tr[2], dr[2]) - g.precision)
  }

use.smoothing.qq <- TRUE


### F difference
w1.fdiff <- function(X, theta, delta, t)
  krnl((t - X) / bw.X) - theta

w2.fdiff <- function(X, theta, delta, t)
  krnl((t - X) / bw.Y) - theta - delta

alpha1.fdiff <- function(X, theta, delta, t)
  -1

alpha2.fdiff <- function(X, theta, delta, t)
  -1

trange.fdiff <- function(X, Y)
  c(max(c(min(X1), min(X2))), min(c(max(X1), max(X2))))

deltarange.fdiff <- function(X, Y, t)
{
  ww <- w1.fdiff(X, 0, 0, t)
  tr <- c(min(ww), max(ww))
  ww <- w2.fdiff(Y, 0, 0, t)
  c(min(ww) - tr[2]+1e-12, max(ww) - tr[1]-1e-12)
}

thetarange.fdiff <- function(X, Y, delta, t)
{
  ww <- w1.fdiff(X, 0, 0, t)
  tr <- c(min(ww)+1e-12, max(ww)-1e-12)
  ww <- w2.fdiff(Y, 0, delta, t)
  tr2 <- c(min(ww)+1e-12, max(ww)-1e-12)
  c(max(tr2[1], tr[1]), min(tr2[2], tr[2]))
}

use.smoothing.fdiff <- TRUE


### Quantile difference

w1.qdiff <- function(X, theta, delta, t)
  krnl((theta - X) / bw.X) - t

w2.qdiff <- function(X, theta, delta, t)
  krnl((theta + delta - X) / bw.Y) - t

alpha1.qdiff <- function(X, theta, delta, t)
  dkrnl((theta - X) / bw.X) / bw.X

alpha2.qdiff <- function(X, theta, delta, t)
  dkrnl((theta + delta - X) / bw.Y) / bw.Y

trange.qdiff <- function(X, Y)
  {
    minlen <- min(length(X), length(Y))
    c(1/minlen, 1 - 1/minlen)
  }

deltarange.qdiff <- function(X, Y, t)
  {
    XX <- sort(X)
    YY <- sort(Y)
    nX <- length(X)
    nY <- length(Y)
    c(YY[2] - XX[nX - 1] + g.precision, YY[nY - 1] - XX[2] - g.precision)
  }

thetarange.qdiff <- function(X, Y, delta, t)
  {
    XX <- sort(X)
    YY <- sort(Y)
    nX <- length(X)
    nY <- length(Y)
    dX <- bw.X * ikrnl(1-t)
    dY <- bw.Y * ikrnl(1-t)
    c(max(XX[2] - dX, YY[2] - delta - dY) + g.precision, min(XX[nX - 1] - dX, YY[nY - 1] - delta - dY) - g.precision)
  }

use.smoothing.qdiff <- TRUE


#### Generalized calling

EL.local <- function(fun, ..., bw.fixed=NULL, bw.fun=NULL, bw.exp=-0.2, method="mean")
  {
    local({
      set.method(method)
      set.globals()
      set.kernel()
      if (is.null(bw.fun))
        {
          if (is.null(bw.fixed))
            set.bw("e", bw.exp)
          else
            set.bw("m", bw.fixed)
        }
      else
        {
          set.bw("f", bw.fun)
        }
      fun(...)
    })
  }

#### Calculations

ELconf <- function(X, Y, t, p.level=0.95, delta=NULL)
  {
    initbw(X, Y)
    if(is.null(delta))
      delta <- deltasolve(X, Y, t)

    critval <- qchisq(p.level, 1)
    ELlim <- function(delt)
      {
        ELLR(X, Y, delt, t) - critval
      }

    drange <- deltarange(X, Y, t)

    if (ELlim(delta) < 0)
      {
        if (ELlim(drange[1]) > 0)
          {
            lo <- uniroot(ELlim, c(drange[1], delta))$root
          }
        else
          {
            warning(paste("Could not find lower confidence bound for t =", t, "."))
            lo <- drange[1]
          }
        
        if (ELlim(drange[2]) > 0)
          {
            hi <- uniroot(ELlim, c(delta, drange[2]))$root
          }
        else
          {
            warning(paste("Could not find upper confidence bound for t =", t, "."))
            hi <- drange[2]
          }
      }
    else
      {
        warning(paste("Value out of bounds at t =", t, "."))
        hi <- delta
        lo <- delta
      }
      
    c(lo, hi)
  }

deltasolve <- function(X, Y, t)
  {
    initbw(X, Y)
    dr <- deltarange(X, Y, t)
    opt <- Inf
    oo <- optimize(function(d) ELLR(X, Y, d, t), dr)
    optnew <- oo$objective
    mino <- oo$minimum
    #return(optnew)
    i <- 0
    while (opt > optnew + 0.1)
	{ 
        #if (i > 0) print(paste(opt, optnew))
        i <- i+1
    	  opt <- optnew
        o1 <- optimize(function(d) ELLR(X, Y, d, t), c(dr[1], mino-0.001))
        o2 <- optimize(function(d) ELLR(X, Y, d, t), c(mino+0.001, dr[2]))
        if (o1$objective < o2$objective)
          {
             mino <- o1$minimum
             optnew <- o1$objective
          }
        else
          {
             mino <- o2$minimum
             optnew <- o2$objective
          }
	}
    mino
  }

ELLR <- function(X, Y, delta, t)
  {
    initbw(X, Y)
    theta <- thetasolve(X, Y, delta, t)
    if (!g.degenerate)
      2* (sum(log(1 + g.lambda1 * g.wval1)) +
          sum(log(1 + g.lambda2 * g.wval2)))
    else
      Inf
  }

#### Internal functions

thetasolve <- function(X, Y, delta, t)
  {
	initbw(X1, X2)
    tgrad <- function(theta)
      {
        g.degenerate <<- FALSE
        g.wval1 <<- w1(X, theta, delta, t)
        g.wval2 <<- w2(Y, theta, delta, t)
        g.lambdarange1 <<- c(-1/max(g.wval1)+ 1e-7,
                             -1/min(g.wval1)- 1e-7)
	  if ((max(g.wval1) == 0) | (min(g.wval1) == 0))
	    {
	      g.degenerate <<- TRUE
		if (max(g.wval1) == 0)
		  return(sign(min(g.wval1)) * -Inf)
		else
 		  return(sign(max(g.wval1)) * -Inf)	
	    }
        if (sign(g.lambdarange1[1]) == sign(g.lambdarange1[2]))
          {
            g.degenerate <<- TRUE
            return(sign(g.lambdarange1[1]) * Inf)
          }
        g.lambdarange2 <<- c(-1/max(g.wval2)+1e-7,
                             -1/min(g.wval2)-1e-7)
	  if ((max(g.wval2) == 0) | (min(g.wval2) == 0))
	    {
	      g.degenerate <<- TRUE
		if (max(g.wval2) == 0)
		  return(sign(min(g.wval2)) * -Inf)
		else
 		  return(sign(max(g.wval2)) * -Inf)	
	    }
        if (sign(g.lambdarange2[1]) == sign(g.lambdarange2[2]))
          {
            g.degenerate <<- TRUE
            return(sign(g.lambdarange2[1]) * Inf)
          }
        alphaval1 <- alpha1(X, theta, delta, t)
        alphaval2 <- alpha2(Y, theta, delta, t)
        g.lambda1 <<- lambdasolve1()
        g.lambda2 <<- lambdasolve2()
        return(
               g.lambda1 * sum(alphaval1 / (1 + g.lambda1 * g.wval1)) +
               g.lambda2 * sum(alphaval2 / (1 + g.lambda2 * g.wval2)))
      }

    suppressWarnings(try(uniroot(tgrad, thetarange(X, Y, delta, t))$root, silent=TRUE))
  }

 
lambdasolve1 <- function()
  {
    uniroot(function(l) sum(g.wval1 / (1 + l * g.wval1)), g.lambdarange1)$root
  }

lambdasolve2 <- function()
  {
    uniroot(function(l) sum(g.wval2 / (1 + l * g.wval2)), g.lambdarange2)$root
  }

